# Copyright (c) AffectLog SAS
# Licensed under the MIT License.

import atexit
import logging
import socket
import threading
import time
import uuid
import warnings

from flask import Flask
from flask_socketio import SocketIO
from gevent.pywsgi import WSGIServer

from .environment_detector import build_environment
from .environments.credentialed_vm_environment import CREDENTIALED_VM
from .environments.public_vm_environment import PUBLIC_VM

LOCALHOST = 'localhost'
VM_ENVS = {CREDENTIALED_VM, PUBLIC_VM}


class FlaskHelper(object):
    """FlaskHelper is a class for common Flask utilities used in dashboards."""

    def __init__(self, ip=None, port=None, with_credentials=False,
                 is_private_link=False):
        # The name passed to Flask needs to be unique per instance.
        self.app = Flask(uuid.uuid4().hex)

        self.port = port
        self.ip = ip
        self.with_credentials = with_credentials
        self.is_private_link = is_private_link
        # dictionary to store arbitrary state for use by consuming classes
        self.shared_state = {}
        if self.ip is None:
            self.ip = "localhost"
        if self.port is None:
            # Try 100 different ports
            available = False
            for port in range(8704, 8994):
                available = FlaskHelper._is_local_port_available(
                    self.ip, port, taise_error=False)
                if available:
                    self.port = port
                    break

            if not available:
                error_message = """Ports 8704 to 8993 not available.
                    Please specify an open port for use via the 'port'
                    parameter"""
                raise RuntimeError(
                    error_message.format(port)
                )
        else:
            FlaskHelper._is_local_port_available(self.ip, self.port,
                                                 taise_error=True)
        self.env = build_environment(self)
        if hasattr(self.env, 'nbvm_origins'):
            self.socketio = SocketIO(
                self.app, cors_allowed_origins=self.env.nbvm_origins)
        else:
            self.socketio = SocketIO(self.app)
        if self.env.base_url is None:
            return
        # Sleep for 1 second in order to prevent random errors while
        # socket is still closing
        time.sleep(1)
        self._thread = threading.Thread(target=self.run, daemon=True)
        self._thread.start()

        # Closes server on program exit, including freeing all sockets
        def closeserver():
            self.stop()

        atexit.register(closeserver)

    @staticmethod
    def _is_local_port_available(ip, port, taise_error=True):
        """Check whether the specified local port is available.

        Borrowed from:
        https://stackoverflow.com/questions/19196105/how-to-check-if-a-network-port-is-open-on-linux
        """
        try:
            backlog = 5
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
                # See stack overflow to prevent "Only one usage" random
                # errors in tests:
                # https://stackoverflow.com/questions/30420512/python-socket-error-only-one-usage-of-each-socket-address-is-normally-permitted
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
                sock.bind((LOCALHOST, port))
                sock.listen(backlog)
        except (socket.error, OSError):  # pragma: no cover
            if taise_error:
                error_message = """Port {0} is not available.
                Please specify another port for use via the 'port' parameter"""
                raise RuntimeError(
                    error_message.format(port)
                )
            else:
                return False
        return True

    def run(self):
        ip = LOCALHOST
        # Note: for credentialed or public VM use the private IP address
        if self.env_name in VM_ENVS:
            host_name = socket.gethostname()
            ip = socket.gethostbyname(host_name)
        logger = logging.getLogger('wsgiserver')
        logger.setLevel(logging.ERROR)
        self.server = WSGIServer((ip, self.port), self.app, log=logger)
        self.app.config["server"] = self.server
        self.socketio.run(self.app, host=ip, port=self.port)
        self.server.serve_forever()

    def stop(self):
        if (self.server.started):
            try:
                self.server.stop()
            except Exception as e:
                warning_msg = "Caught exceptions when closing server: {}"
                warning_msg = warning_msg.format(e)
                warnings.warn(warning_msg, UserWarning)



# Copyright (c) AffectLog SAS
# Licensed under the MIT License.

"""Module for environments for flask."""

from .azure_nb_environment import AzureNBEnvironment
from .credentialed_vm_environment import CredentialedVMEnvironment
from .databricks_environment import DatabricksEnvironment
from .local_ipython_environment import LocalIPythonEnvironment
from .public_vm_environment import PublicVMEnvironment

__all__ = [
    'AzureNBEnvironment',
    'CredentialedVMEnvironment',
    'DatabricksEnvironment',
    'LocalIPythonEnvironment',
    'PublicVMEnvironment'
]



# Copyright (c) AffectLog SAS
# Licensed under the MIT License.

import os

import pytest

from tai_core_flask import FlaskHelper
from tai_core_flask.environments.azure_nb_environment import AzureNBEnvironment
from tai_core_flask.environments.credentialed_vm_environment import \
    CredentialedVMEnvironment
from tai_core_flask.environments.databricks_environment import (
    DATABRICKS_ENV_VAR, DatabricksEnvironment)
from tai_core_flask.environments.local_ipython_environment import \
    LocalIPythonEnvironment
from tai_core_flask.environments.public_vm_environment import \
    PublicVMEnvironment


class TestEnvironmentDetector(object):

    def test_credentialed_vm(self):
        service = FlaskHelper(with_credentials=True)
        assert isinstance(service.env, CredentialedVMEnvironment)

    def test_public_vm(self, mocker):
        mocker.patch('tai_core_flask.FlaskHelper._is_local_port_available',
                     return_value=True)
        service = FlaskHelper(ip="not localhost", with_credentials=False)
        assert isinstance(service.env, PublicVMEnvironment)

    def test_public_vm_fail_on_port(self, mocker):
        mocker.patch('tai_core_flask.FlaskHelper._is_local_port_available',
                     return_value=False)
        with pytest.raises(RuntimeError) as exception:
            FlaskHelper(ip="not localhost", with_credentials=False)
        assert "Ports 8704 to 8993 not available." in \
            exception.value.args[0]

    def test_local(self):
        service = FlaskHelper()
        assert isinstance(service.env, LocalIPythonEnvironment)

    def test_azure_nb(self, mocker):
        mocker.patch('os.path.exists', return_value=True)
        mocker.patch('os.path.isfile', return_value=True)
        mocker.patch(
            'tai_core_flask.environments.azure_nb_environment.'
            'AzureNBEnvironment.get_nbvm_config',
            return_value={
                'instance': "fakeaznbinstance",
                'domainsuffix': "fakedomainsuffix"})
        service = FlaskHelper()
        assert isinstance(service.env, AzureNBEnvironment)
        assert service.with_credentials

    def test_databricks(self):
        try:
            os.environ[DATABRICKS_ENV_VAR] = "mock"
            service = FlaskHelper()
            assert isinstance(service.env, DatabricksEnvironment)
        finally:
            del os.environ[DATABRICKS_ENV_VAR]



# Copyright (c) AffectLog SAS
# Licensed under the MIT License.

import abc


class BaseEnvironment(abc.ABC):
    """Interface for environments."""
    def display(self, html):
        from IPython.display import HTML, display
        display(HTML(html), metadata=dict(isolated=True))

    def select(self, service):
        pass



# Copyright (c) AffectLog SAS
# Licensed under the MIT License.

from flask_cors import CORS

from tai_core_flask.environments.base_environment import BaseEnvironment

CREDENTIALED_VM = 'credentialed_vm'


class CredentialedVMEnvironment(BaseEnvironment):
    """Environment class for credentialed VM environments.

    CredentialedVMEnvironment represents functionality to detect whether it is
    executed in a credentialed VM environment based on the with_credentials
    parameter. Additionally, it can display corresponding visualizations.
    """

    def __init__(self, service):
        self.successfully_detected = False
        self.base_url = None

        if service.with_credentials:
            self.successfully_detected = True

    def select(self, service):
        origin = f"https://{service.ip}:{service.port}"
        headers = ['Content-Type']
        service.cors = CORS(service.app, origins=[origin],
                            expose_headers=headers, supports_credentials=True)
        service.env_name = CREDENTIALED_VM



# Copyright (c) AffectLog SAS
# Licensed under the MIT License.

import os

from .environments import (AzureNBEnvironment, CredentialedVMEnvironment,
                           DatabricksEnvironment, LocalIPythonEnvironment,
                           PublicVMEnvironment)

"""Environment detection related utilities.
A good portion of this code has largely been sourced from open-source licensed
code available between StackOverflow and plotly.

Plotly derived code comes from below:
https://github.com/plotly/plotly.py/blob/944af4a0b28bef2b139307a7808c02c51d804c4d/packages/python/plotly/plotly/io/_renderers.py#L455
"""


def _detect_ipython():
    """Detect if called in an IPython environment.

    Mostly derived from StackOverflow below:
    https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook

    Returns
    -------
        True if in IPython environment.

    """
    try:
        from IPython import get_ipython

        return get_ipython() is not None
    except NameError:  # pragma: no cover
        return False


def _detect_ipython_zmq():
    """Detect if running in an IPython environment using ZMQ.

    i.e. notebook/qtconsole/lab.

    Mostly derived from StackOverflow below:
    https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook/24937408

    Returns
    -------
        True if called in IPython notebook or qtconsole.

    """
    try:
        from IPython import get_ipython

        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:  # pragma: no cover
        return False  # Probably standard Python interpreter


def _detect_colab():
    try:
        import google.colab  # noqa: F401

        return True
    except ImportError:
        return False


def _detect_binder():
    return "BINDER_SERVICE_HOST" in os.environ


def _detect_sagemaker():
    return "SM_NUM_CPUS" in os.environ


def _detect_kaggle():
    return os.path.exists("/kaggle/input")


def _detect_azure_notebook():
    return "AZURE_NOTEBOOKS_HOST" in os.environ


def _detect_azureml_notebook_vm():
    return "AZUREML_NB_PATH" in os.environ


def _detect_vscode():
    return "VSCODE_PID" in os.environ


def _detect_databricks():
    return "DATABRICKS_RUNTIME_VERSION" in os.environ


def is_cloud_env(detected):
    """Detect whether it is running in a cloud environment."""
    cloud_env = [
        "databricks",
        "azure",
        "azureml_vm",
        "kaggle",
        "sagemaker",
        "binder",
        "colab",
    ]
    if len(set(cloud_env).intersection(detected)) != 0:
        return True
    else:
        return False


def build_environment(service):
    """Create a runtime environment object."""
    # legacy
    checks = {  # noqa: F841
        "azure_nb": AzureNBEnvironment,
        "databricks": _detect_databricks,
        "vscode": _detect_vscode,
        "azure": _detect_azure_notebook,
        "azureml_vm": _detect_azureml_notebook_vm,
        "kaggle": _detect_kaggle,
        "sagemaker": _detect_sagemaker,
        "binder": _detect_binder,
        "colab": _detect_colab,
        "ipython-zmq": _detect_ipython_zmq,
        "ipython": _detect_ipython,
    }

    azure_nb_environment = AzureNBEnvironment(service)
    credentialed_vm_environment = CredentialedVMEnvironment(service)
    databricks_environment = DatabricksEnvironment(service)
    local_ipython_environment = LocalIPythonEnvironment(service)
    public_vm_environment = PublicVMEnvironment(service)

    # Todo: Add case for detecting and using a Jupyterlab environment
    # note: the order matters since Databricks requirements
    # are a superset of local requirements, for example
    if databricks_environment.successfully_detected:
        env = databricks_environment
    elif azure_nb_environment.successfully_detected:
        env = azure_nb_environment
    elif local_ipython_environment.successfully_detected:
        env = local_ipython_environment
    elif public_vm_environment.successfully_detected:
        env = public_vm_environment
    elif credentialed_vm_environment.successfully_detected:
        env = credentialed_vm_environment
    else:
        raise Exception("Failed to detect Ipython environment")

    env.select(service)  # TODO: simplify
    return env



# Copyright (c) AffectLog SAS
# Licensed under the MIT License.

import os

from tai_core_flask.environments.base_environment import BaseEnvironment

_DISPLAY_HTML = "displayHTML"
_DISPLAY = "display"
_SPARK = "spark"

DATABRICKS_ENV_VAR = "DATABRICKS_RUNTIME_VERSION"

DATABRICKS = "databricks"


class DatabricksEnvironment(BaseEnvironment):
    """Environment class for Databricks environments.

    DatabricksEnvironment represents functionality to detect whether it is
    executed in a Databricks environment based on environment variables.
    Additionally, it can display corresponding visualizations.
    """

    lazy_display_function = None

    def __init__(self, service):
        self.successfully_detected = False
        self.base_url = None
        self.with_credentials = False

        if DATABRICKS_ENV_VAR in os.environ:
            self.successfully_detected = True

    def display(self, html):
        """Display the passed HTML using Databricks's displayHTML.

        ..note:

            Code mostly derived from Plotly's databricks render as linked
            below:
            https://github.com/plotly/plotly.py/blob/01a78d3fdac14848affcd33ddc4f9ec72d475232/packages/python/plotly/plotly/io/_base_renderers.py
        """
        import inspect

        if self.lazy_display_function is None:
            found = False
            for frame in inspect.getouterframes(inspect.currentframe()):
                global_names = set(frame.frame.f_globals)
                target_names = {_DISPLAY_HTML, _DISPLAY, _SPARK}
                if target_names.issubset(global_names):
                    self.lazy_display_function = \
                        frame.frame.f_globals[_DISPLAY_HTML]
                    found = True
                    break

            if not found:
                msg = "Could not find Databrick's displayHTML function"
                raise RuntimeError(msg)

        self.lazy_display_function(html)

    def select(self, service):
        service.env_name = DATABRICKS



# Copyright (c) AffectLog SAS
# Licensed under the MIT License.

import time

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from tai_core_flask import FlaskHelper


class TestFlaskHelper(object):

    def setup_class(cls):
        """Set up once for all tests."""
        cls.port_range_start = 8704
        cls.port_range_end = 8993
        cls.test_ip_local = "localhost"
        cls.test_port_local = 8994

    def get_http_client(self):
        """Get HTTP client with automatic retries."""
        retry_strategy = Retry(total=3)
        adapter = HTTPAdapter(max_retries=retry_strategy)
        http = requests.Session()
        http.mount("https://", adapter)
        http.mount("http://", adapter)

        return http

    def test_without_explicit_port(self):
        """Test the flask helper without setting an explicit port."""
        http = self.get_http_client()

        flask_service = FlaskHelper(ip=self.test_ip_local)

        assert (flask_service.port >= 8704 and flask_service.port <= 8993)

        @flask_service.app.route("/hello", methods=["GET"])
        def hello():
            return "Hello"

        response = call_get_with_retries(
            http, f"{flask_service.env.base_url}/hello")

        assert (response.status_code == 200)
        assert (response.text == "Hello")

    def test_with_explicit_port(self):
        """Test the flask helper when setting an explicit port."""
        http = self.get_http_client()

        flask_service = FlaskHelper(ip=self.test_ip_local,
                                    port=self.test_port_local)

        assert (flask_service.port == self.test_port_local)

        @flask_service.app.route("/hello_two", methods=["GET"])
        def hello_two():
            return "Hello"

        response = call_get_with_retries(
            http, f"{flask_service.env.base_url}/hello_two")

        assert (response.status_code == 200)
        assert (response.text == "Hello")


def call_get_with_retries(http_client,
                          url_route,
                          max_retries=4,
                          retry_delay=60):
    """Call http get on route with retries until successful."""
    for i in range(max_retries):
        try:
            print("Calling get attempt {0} of {1}".format(i + 1, max_retries))
            response = http_client.get(url_route)
            break
        except Exception as e:  # noqa: B902
            print("Calling get attempt failed with exception:")
            print(e)
            if i + 1 != max_retries:
                print("Will retry after {0} seconds".format(retry_delay))
                time.sleep(retry_delay)
                retry_delay = retry_delay * 2
    else:
        raise RuntimeError("Unable to call get on the given url route")

    return response



# Copyright (c) AffectLog SAS
# Licensed under the MIT License.

import os
import re

from flask_cors import CORS

from tai_core_flask.environments.base_environment import BaseEnvironment

AZURE_NB = "azure_nb"


def get_url_format(is_private_link, instance_name, domain_suffix, port):
    if is_private_link:
        return f"{instance_name}.{domain_suffix}:{port}"
    return f"{instance_name}-{port}.{domain_suffix}"


class AzureNBEnvironment(BaseEnvironment):
    """Environment class for Azure notebook environments.

    AzureNBEnvironment represents functionality to detect whether it is
    executed in a Azure notebook environment based on the `.nbvm` file.
    Additionally, it can display corresponding visualizations.
    """

    nbvm_file_path = "/mnt/azmnt/.nbvm"

    def __init__(self, service):
        self.base_url = None
        self.successfully_detected = False

        if not (os.path.exists(AzureNBEnvironment.nbvm_file_path) and
                os.path.isfile(AzureNBEnvironment.nbvm_file_path)):
            self.successfully_detected = False
        else:
            self.nbvm = self.get_nbvm_config()

            if "instance" not in self.nbvm or "domainsuffix" not in self.nbvm:
                self.successfully_detected = False
            else:
                instance_name = self.nbvm["instance"]
                domain_suffix = self.nbvm["domainsuffix"]
                url_format = get_url_format(
                    service.is_private_link, instance_name, domain_suffix,
                    service.port)
                self.base_url = f"https://{url_format}"
                self.successfully_detected = True
                self.nbvm_origins = [
                    f"https://{instance_name}.{domain_suffix}",
                    f"https://{url_format}",
                    f"wss://{instance_name}.{domain_suffix}",
                    f"wss://{url_format}",
                    "https://ml.azure.com",
                    "https://dev.ml.azure.com",
                    "https://int.ml.azure.com"
                ]

    def get_nbvm_config(self):
        # Use regex to find items of the form key=value where value is a
        # part of a URL.
        # The keys of interest are "instance" and domainsuffix"
        envre = re.compile(r'''^([^\s=]+)=(?:[\s"']*)(.+?)(?:[\s"']*)$''')
        result = {}
        with open(AzureNBEnvironment.nbvm_file_path) as nbvm_variables:
            for line in nbvm_variables:
                match = envre.match(line)
                if match is not None:
                    result[match.group(1)] = match.group(2)
        return result

    def select(self, service):
        headers = ['Content-Type']
        service.cors = CORS(
            service.app,
            origins=self.nbvm_origins,
            expose_headers=headers,
            supports_credentials=True)
        service.with_credentials = True
        service.env_name = AZURE_NB



# Copyright (c) AffectLog SAS
# Licensed under the MIT License.

from flask_cors import CORS

from tai_core_flask.environments.base_environment import BaseEnvironment

LOCAL = 'local'


class LocalIPythonEnvironment(BaseEnvironment):
    """Environment class for local IPython environments.

    LocalIPythonEnvironment represents functionality to detect whether it is
    executed in a local python environment based on IPython's availability.
    Additionally, it can display corresponding visualizations.
    """

    def __init__(self, service):
        self.successfully_detected = False
        self.base_url = None

        try:
            from IPython.display import HTML, display  # noqa: F401
        except NameError:
            self.successfully_detected = False
        else:
            if service.ip == "localhost" and not service.with_credentials:
                self.successfully_detected = True
                self.base_url = f"http://localhost:{service.port}"

    def select(self, service):
        service.with_credentials = False
        service.cors = CORS(service.app)
        service.env_name = LOCAL



# Copyright (c) AffectLog SAS
# Licensed under the MIT license.

import setuptools

# The version must be incremented every time we push an update to pypi (but
# not before)
VERSION = "0.7.7"

# supply contents of our README file as our package's long description
with open("README.md", "r") as fh:
    long_description = fh.read()


requirements = []
with open("requirements.txt", "r") as fr:
    requirements = list(filter(
        lambda rq: rq != "",
        map(lambda r: r.strip(), fr.read().split("\n"))))


setuptools.setup(
    # this is the name people will use to "pip install" the package
    name="tai_core_flask",

    version=VERSION,
    author="AL360°",
    author_email="developer@affectlog.com",
    description="Trustworthy AI Core Flask Wrapper",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    # this will find our package "xtlib" by its having an "__init__.py" file
    packages=setuptools.find_packages(),
    entry_points={},
    # normally, only *.py files are included - this forces our YAML file and
    # controller scripts to be included
    package_data={'': []},
    include_package_data=True,
    # the packages that our package is dependent on
    install_requires=requirements,
    extras_require=dict(
        dev=[
        ], ),
    # used to identify the package to various searches
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
